Require Import Relations.

Require Import Wf.
Require Import Wellfounded.

Set Implicit Arguments.

Section Newman.

  Variables (X : Type).

  Section defs.

    Implicit Type R : X -> X -> Prop.
    
    Definition diamond R := forall x a b, R a x -> R b x -> exists y, R y a /\ R y b.
    Definition localy_confluent R := forall x a b, R a x -> R b x -> exists y, clos_refl_trans _ R y a /\ clos_refl_trans _ R y b.
    Definition confluent R := diamond (clos_refl_trans _ R).

  End defs.
 
  Variable R : X -> X -> Prop.

  Hypothesis Rwconf : localy_confluent R.
  Hypothesis Rwf : well_founded R.

  Proposition clos_t_rt_t x y z : clos_trans _ R x y -> clos_refl_trans _ R y z -> clos_trans _ R x z.
  Proof.
    intros H1 H2; revert H2 x H1.
    induction 1.
    intros ? Hz; constructor 2 with (1 := Hz); constructor 1; auto.
    intros; auto.
    intros; auto.
  Qed.
  
  Proposition clos_rt_t_t x y z : clos_refl_trans _ R x y -> clos_trans _ R y z -> clos_trans _ R x z.
  Proof.
    intros H1; revert H1 z.
    induction 1.
    intros ? Hz; constructor 2 with (2 := Hz); constructor 1; auto.
    intros; auto.
    intros; auto.
  Qed.

  Lemma Newmans_Lemma : confluent R.
  Proof.    
    intros x.
    induction x as [ x IHx ] using (well_founded_induction (wf_clos_trans _ _ Rwf)); intros a b Ha Hb.
    apply clos_rt_rtn1_iff in Ha.
    destruct Ha as [ | a' x Ha' ].
    exists b; split; auto; constructor 2.
    apply clos_rt_rtn1_iff in Hb.
    destruct Hb as [ | b' x Hb' ].
    exists a; split.
    constructor 2.
    apply clos_rt_rtn1_iff.
    apply Relation_Operators.rtn1_trans with a'; auto.
    destruct (Rwconf Ha' Hb') as (y & H1 & H2).
    apply clos_rt_rtn1_iff in Ha.
    apply clos_rt_rtn1_iff in Hb.
    destruct (IHx a') with (a := a) (b := y) as ( yl & H3 & H4 ); auto. constructor 1; auto.
    destruct (IHx b') with (a := y) (b := b) as ( yr & H5 & H6 ); auto. constructor 1; auto.
    destruct (IHx y) with (a := yl) (b := yr) as (yc & H7 & H8); auto.
    apply clos_rt_t_t with (1 := H1); constructor 1; auto.
    exists yc; split.
    apply rt_trans with (1 := H7); auto.
    apply rt_trans with (1 := H8); auto.
  Qed.

End Newman.
    
  
  
  
  