(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Relations Arith Omega.

Require Import utils finitary_rel pos cr.
Require Import uterm.

Set Implicit Arguments.

Reserved Notation " x '-b->' y " (at level 71).

(* Beware that ut_beta x y is denote y -b-> x (reversed) *)

Inductive ut_beta : forall n, uterm n -> uterm n -> Prop :=
  | in_ut_beta_redex : forall n f (a   : uterm n),              \\ f @ a -b-> f[a]  
  | in_ut_beta_lft   : forall n (x y k : uterm n),     x -b-> y -> x @ k -b-> y @ k
  | in_ut_beta_rt    : forall n (x y k : uterm n),     x -b-> y -> k @ x -b-> k @ y
  | in_ut_beta_abs   : forall n (x y : uterm (S n)),   x -b-> y -> \\ x  -b-> \\ y
where  " x '-b->' y " := (@ut_beta _ x y).

Fact ut_beta_id n p : ut_id n @ p -b-> p.
Proof.
  unfold ut_id; simpl.
  cutrewrite (p = £(pos_fst n) [p]); auto.
  apply in_ut_beta_redex.
Qed.

Local Notation  " x '~[' R ']~' y " := (@ut_ren_rel _ _ R x y) (at level 71).

Section ut_beta_ren_rel_inj.

  Fact ut_beta_ren_rel_fun n m (R : ren_rel n m) fn gn fm gm :
       (forall x y1 y2, R x y1 -> R x y2 -> y1 = y2)
    -> fn ~[R]~ fm -> gn ~[R]~ gm -> fn -b-> gn -> fm -b-> gm.
  Proof.
    intros HR H1 H2 H; revert H m fm gm R HR H1 H2.
  
    induction 1 as [ n f' a' 
                   | n f' h' a' H1 IH 
                   | n a' b' f' H1 IH 
                   | n f' h' H1 IH1 ]; intros m fm gm R HR H3 H4.
                 
    apply ut_ren_rel_app_inv in H3.
    destruct H3 as (f & a & ? & H1 & H2); subst.
    apply ut_ren_rel_abs_inv in H1.
    destruct H1 as (f'' & ? & H1); subst.
    generalize (ut_ren_rel_reduce H1 H2); intros H3.
    rewrite (ut_ren_rel_fun HR H4 H3).
    constructor 1.
  
    apply ut_ren_rel_app_inv in H3.
    destruct H3 as (f & a & ? & H5 & H6); subst.
    apply ut_ren_rel_app_inv in H4.
    destruct H4 as (h & b & ? & H7 & H8); subst.
    generalize (ut_ren_rel_fun HR H6 H8); intro; subst b; clear H8.
    constructor 2; apply IH with (1 := HR); auto.
      
    apply ut_ren_rel_app_inv in H3.
    destruct H3 as (f & b & ? & H5 & H6); subst.
    apply ut_ren_rel_app_inv in H4.
    destruct H4 as (h & c & ? & H7 & H8); subst.
    generalize (ut_ren_rel_fun HR H5 H7); intro; subst h; clear H7.
    constructor 3; apply IH with (1 := HR); auto.
    
    apply ut_ren_rel_abs_inv in H3.
    destruct H3 as (f & ? & H3); subst.
    apply ut_ren_rel_abs_inv in H4.
    destruct H4 as (h & ? & H4); subst.
    constructor 4; apply IH1 with (2 := H3); auto.
    intros p u v; pos_inv p; pos_inv u; pos_inv v; try (destruct 1; fail); try (destruct 2; fail); auto.
    unfold ren_rel_lift; simpl; intros; f_equal; apply HR with p; auto.
  Qed.
  
End ut_beta_ren_rel_inj.

Notation ut_unused := ((fun n f p => @ut_occ n f p = 0) _).

Fact ut_beta_unused n (f g : uterm n) : f -b-> g -> ut_unused f inc1 ut_unused g.
Proof.
  induction 1 as [ n f' a' 
                 | n f' h' a' H1 IH 
                 | n f' a' b' H1 IH 
                 | n f' h' H1 IH1 ]; intros p; simpl; intros H.
  rewrite ut_occ_reduce.
  apply plus_is_O in H; destruct H as [ H1 H2 ].
  rewrite H1, H2, mult_comm; auto.
  apply plus_is_O in H; destruct H as [ H2 H3 ].
  apply IH in H2; omega.
  apply plus_is_O in H; destruct H as [ H2 H3 ].  
  apply IH in H3; omega.
  auto.
Qed.

Definition ut_beta_rt n := clos_refl_trans _ (@ut_beta n).

Notation " x '-b>>' y " := (@ut_beta_rt _ x y) (at level 71).

Fact ut_beta_rt_unused n (f g : uterm n) : f -b>> g -> ut_unused f inc1 ut_unused g.
Proof.
  induction 1; auto.
  apply ut_beta_unused; auto.
Qed.

Fact ut_beta_rt_lift_not_pos0 n (f : uterm n) : ~ ut_lift f -b>> £ pos0.
Proof.
  intros H.
  generalize (@ut_beta_rt_unused _ _ _ H pos0).
  rewrite ut_occ_lift_fst; simpl; omega.
Qed.

Fact ut_beta_rt_refl n (f : uterm n) : f -b>> f.
Proof. constructor 2. Qed.

Fact ut_beta_rt_refl_eq n (f g : uterm n) : f = g -> f -b>> g.
Proof. intros []; constructor 2. Qed.

Fact ut_eq_beta_rt n (f g : uterm n) : f = g -> f -b>> g.
Proof. intros []; constructor 2. Qed.

Fact ut_beta_rt_trans n (f g h : uterm n) : f -b>> g -> g -b>> h -> f -b>> h.
Proof. intros; constructor 3 with g; auto. Qed.

Fact ut_beta_beta_rt n (f g : uterm n) : f -b-> g -> f -b>> g.
Proof. intros; constructor 1; auto. Qed.

Fact ut_beta_rt_inv n (f g : uterm n) : f -b>> g -> f = g \/ exists h, f -b-> h /\ h -b>> g.
Proof. apply clos_refl_trans_step. Qed.

Fact ut_beta_redex_rt n f (a h : uterm n) : f[a] -b>> h -> \\f @ a -b>> h.
Proof.
  intros H.
  apply ut_beta_rt_trans with (2 := H).
  constructor 1; apply in_ut_beta_redex.
Qed.

Fact ut_beta_rt_lft n (f g a : uterm n) : f -b>> g -> f @ a -b>> g @ a.
Proof.
  apply clos_refl_trans_map with (f := fun x => x @ a).
  intros ? ?; apply in_ut_beta_lft.
Qed.

Fact ut_beta_rt_rt n (f a b : uterm n) : a -b>> b -> f @ a -b>> f @ b.
Proof.
  apply clos_refl_trans_map with (f := fun x => f @ x).
  intros ? ?; apply in_ut_beta_rt.
Qed.

Fact ut_beta_rt_app n (f g a b : uterm n) : f -b>> g -> a -b>> b -> f @ a -b>> g @ b.
Proof.
  intros H1 H2.
  constructor 3 with (f @ b).
  apply ut_beta_rt_rt; auto.
  apply ut_beta_rt_lft; auto.
Qed.

Fact ut_beta_rt_abs n (f g : uterm (S n)) : f -b>> g -> \\ f -b>> \\ g.
Proof.
  apply clos_refl_trans_map with (f := @ut_abs _).
  intros ? ?; apply in_ut_beta_abs.
Qed.

Fact ut_beta_subst n (f g : uterm n) m (s : uts_t n m) : f -b-> g -> ut_subst f s -b-> ut_subst g s.
Proof.
  intros H; revert H m s.
  induction 1 as [ n f' a' | n f' h' a' H1 IH | n f' a' b' H1 IH | n f' h' H IH ]; intros m s; simpl.
  rewrite ut_subst_reduce; apply in_ut_beta_redex.
  apply in_ut_beta_lft; auto.
  apply in_ut_beta_rt; auto.
  apply in_ut_beta_abs; auto.
Qed.

Fact ut_beta_rt_subst n (f g : uterm n) m (s : uts_t n m) : f -b>> g -> ut_subst f s -b>> ut_subst g s.
Proof.
  apply clos_refl_trans_map with (f := fun x => ut_subst x s).
  intros ? ?; apply ut_beta_subst.
Qed.

Fact ut_beta_reduce n (f g : uterm (S n)) a : f -b-> g -> f[a] -b-> g[a].
Proof.
  apply ut_beta_subst.
Qed.

Fact ut_beta_ren n (f g : uterm n) m (r : pos_map n m) : f -b-> g -> ut_ren f r -b-> ut_ren g r.
Proof. do 2 rewrite ut_ren_subst_eq; apply ut_beta_subst. Qed.

Fact ut_beta_rt_ren n (f g : uterm n) m (r : pos_map n m) : f -b>> g -> ut_ren f r -b>> ut_ren g r.
Proof. do 2 rewrite ut_ren_subst_eq; apply ut_beta_rt_subst. Qed.

Fact ut_beta_lift n (f g : uterm n) : f -b-> g -> ut_lift f -b-> ut_lift g.
Proof. apply ut_beta_ren. Qed.

Fact ut_beta_rt_lift n (f g : uterm n) : f -b>> g -> ut_lift f -b>> ut_lift g.
Proof. apply ut_beta_rt_ren. Qed.

Fact ut_subst_beta_rt n (f : uterm n)  m (r s : uts_t n m) :
  (forall p, r p -b>> s p) -> ut_subst f r -b>> ut_subst f s.
Proof.
  revert m r s.
  induction f as [ n f Hf | n a Ha b Hb | n v ]; intros m r s Hrs; simpl; auto.
  apply ut_beta_rt_abs, Hf.
  intros p; pos_inv p; unfold uts_lift; simpl.
  constructor 2.
  apply ut_beta_rt_lift; auto.
  apply ut_beta_rt_app; auto.
Qed.

Fact ut_reduce_beta_rt n (f : uterm (S n)) a b : a -b>> b -> f[a] -b>> f[b].
Proof.
  intros H; apply ut_subst_beta_rt.
  intros p; unfold uts_first; pos_inv p; auto.
  constructor 2.
Qed.

Fact ut_beta_app_rec n l : forall (f g : uterm n), f -b-> g -> f @* l -b-> g @* l.
Proof.
  induction l as [ | x l IHl ]; simpl; auto; intros f g Hfg.
  apply IHl, in_ut_beta_lft; auto.
Qed.

Fact ut_beta_rt_app_rec n l (f g : uterm n) : f -b>> g -> f @* l -b>> g @* l.
Proof.
  apply clos_refl_trans_map with (f := fun x => x @* l).
  intros ? ?; apply ut_beta_app_rec.
Qed.

Fact ut_beta_app_rec_In n l r : forall (f a b : uterm n), a -b-> b -> f @* (l++a::r) -b-> f @* (l++b::r).
Proof.
  induction l as [ | x l IHl ]; simpl; auto; intros f a b H.
  apply ut_beta_app_rec; constructor 3; auto.
Qed.

Fact ut_beta_app_rec_rev n l (f g : uterm n) : f -b-> g -> f *@ l -b-> g *@ l.
Proof.
  do 2 rewrite ut_app_rev_rec_eq.
  apply ut_beta_app_rec.
Qed.

Fact ut_beta_rt_app_rec_rev n l (f g : uterm n) : f -b>> g -> f *@ l -b>> g *@ l.
Proof.
  do 2 rewrite ut_app_rev_rec_eq.
  apply ut_beta_rt_app_rec.
Qed.

Fact ut_beta_abs_rec k n (f g : uterm (k+n)) : f -b-> g -> \\* f -b-> \\* g.
Proof.
  revert f g; induction k as [ | k IHk ]; simpl; intros f g H; auto.
  apply IHk; constructor; auto.
Qed.

Fact ut_beta_pow_rec n l : forall (f g : uterm n), f -b-> g -> l ^* f -b-> l ^* g.
Proof.
  induction l as [ | x l IHl ]; simpl; auto; intros f g Hfg.
  apply IHl, in_ut_beta_rt; auto.
Qed.

Fact ut_beta_pow_rev n l (f g : uterm n) : f -b-> g -> l *^ f -b-> l *^ g.
Proof.
  do 2 rewrite ut_pow_rev_rec_eq.
  apply ut_beta_pow_rec.
Qed.

Fact ut_beta_rt_pow_rev n l (f g : uterm n) : f -b>> g -> l *^ f -b>> l *^ g.
Proof.
  apply clos_refl_trans_map with (f := fun x => l *^ x).
  intros ? ?; apply ut_beta_pow_rev.
Qed.

Fact ut_beta_rt_church n i (v f g : uterm n) : f -b>> g -> ut_church i v f -b>> ut_church i v g.
Proof.
  apply ut_beta_rt_pow_rev.
Qed.

Fact ut_abs_rec_reduce n m (f : uterm (m+n)) (s : uts_t m n) : 
  ut_app_rec_rev (\\* f) (map s (pos_list m)) -b>> ut_subst f (pf_merge s (@uts_id _)).
Proof.
  revert f s.
  induction m as [ | m IHm ]; intros f s; simpl.

  apply ut_eq_beta_rt.
  rewrite <- (ut_subst_id f) at 1.
  apply ut_subst_ext.
  intros p; unfold pf_merge.
  destruct (@pos_split 0 n p) as [ (q & ?) | (q & ?) ]; subst.
  invert pos q.
  auto.

  generalize (IHm (\\ f) (fun p => s (pos_nxt _ p))); intros H.
  apply ut_beta_rt_lft with (a := s (pos_fst _)) in H.
  rewrite map_map.
  apply ut_beta_rt_trans with (1 := H); clear H.
  simpl.
  apply ut_beta_redex_rt.
  apply ut_eq_beta_rt.
  unfold ut_reduce.
  rewrite ut_subst_comp.
  apply ut_subst_ext.
  intros p; invert pos p; simpl.
  unfold uts_first, pf_merge; simpl; auto.
  unfold uts_lift; simpl.
  rewrite ut_lift_subst_eq, ut_subst_comp.
  rewrite pf_merge_pos_nxt.
  match goal with |- _ = ?f => rewrite <- (ut_subst_id f) at 2 end.
  apply ut_subst_ext.
  intros q; unfold uts_first; simpl; auto.
Qed.

Fact ut_beta_rt_app_rec_rev_fall n f l1 l2 : Forall2 (inv (@ut_beta_rt n)) l2 l1 -> f *@ l1 -b>> f *@ l2.
Proof.
  intros H; revert H f.
  induction 1; intros; simpl.
  constructor 2.
  apply ut_beta_rt_app; auto.
Qed.

Fact ut_beta_rt_app_rec_fall n f l1 l2 : Forall2 (inv (@ut_beta_rt n)) l2 l1 -> f @* l1 -b>> f @* l2.
Proof.
  intros H.
  do 2 rewrite ut_app_rec_rev_eq.
  apply ut_beta_rt_app_rec_rev_fall.
  revert H; clear f.
  induction 1; simpl; auto.
  apply Forall2_app; auto.
Qed.

Fact ut_beta_rt_app_rec_lft_rt n f g l1 l2 : f -b>> g -> Forall2 (inv (@ut_beta_rt n)) l2 l1 -> f @* l1 -b>> g @* l2.
Proof.
  intros H1 H2.
  apply ut_beta_rt_trans with (f @* l2).
  apply ut_beta_rt_app_rec_fall; auto.
  apply ut_beta_rt_app_rec; auto.
Qed.

Fact ut_beta_lift_rec n m (f g : uterm n) : f -b-> g -> ut_lift_rec m f -b-> ut_lift_rec m g.
Proof.
  revert n f g; induction m as [ | m IHm ]; intros n f g; simpl; auto.
  intros H; apply ut_beta_lift, IHm; auto.
Qed.
  
Fact ut_beta_rt_lift_rec n m (f g : uterm n) : f -b>> g -> ut_lift_rec m f -b>> ut_lift_rec m g.
Proof.
  apply clos_refl_trans_map with (f := ut_lift_rec m).
  intros ? ?; apply ut_beta_lift_rec.
Qed.

Fact ut_church_beta n k f (a b : uterm n) : a -b-> b -> ut_church k f a -b-> ut_church k f b.
Proof.
  intros H.
  induction k.
  do 2 rewrite ut_church_O; auto.
  do 2 rewrite ut_church_S_head.
  constructor 3; auto.
Qed.

Fact ut_church_beta_rt n k (f g a b : uterm n) : f -b>> g -> a -b>> b -> ut_church k f a -b>> ut_church k g b.
Proof.
  intros H1 H2.
  apply ut_beta_rt_trans with (ut_church k f b).
  revert H2; apply clos_refl_trans_map with (f := ut_church k f ), ut_church_beta.
  clear a H2.
  induction k as [ | k IHk ].
  do 2 rewrite ut_church_O; apply ut_beta_rt_refl.
  do 2 rewrite ut_church_S_head.
  apply ut_beta_rt_app; auto.
Qed.

Section ut_beta_inv.

  Let ut_beta_inv n (u v : uterm n) : u -b-> v -> (exists f a,   u = \\ f @ a /\ v = f[a])
                                               \/ (exists f g a, u = f @ a    /\ v = g @ a /\ f -b-> g)
                                               \/ (exists f a b, u = f @ a    /\ v = f @ b /\ a -b-> b)
                                               \/ (exists f g,   u = \\ f     /\ v = \\ g  /\ f -b-> g).
  Proof.
    destruct 1 as [ n f a | n f g a H | n a b f H | n f g H ].
    do 0 right; left; exists f, a; auto.
    do 1 right; left; exists f, g, a; auto.
    do 2 right; left; exists f, a, b; auto.
    do 3 right; exists f, g; auto.
  Qed.

  Fact ut_beta_var_inv n (p : pos n) h : ~ £ p -b-> h.
  Proof. inversion 1. Qed.
  
  Fact ut_beta_abs_inv n u (v : uterm n) : \\ u -b-> v -> exists w, v = \\ w /\ u -b-> w.
  Proof.
    intros H; apply ut_beta_inv in H.
    destruct H as [ (f & a & H1 & H2) | [ (f & g & a & H1 & H2 & H3) | [ (f & a & b & H1 & H2 & H3) | (f & g & H1 & H2 & H3) ] ] ]; try discriminate H1.
    apply ut_abs_inj in H1; subst u v; exists g; auto.
  Qed.

  Fact ut_beta_app_inv n (h u v : uterm n) : 
           h @ u -b-> v 
       -> (exists f, h = \\ f /\ v = f[u])
       \/ (exists g, v = g @ u /\ h -b-> g)
       \/ (exists w, v = h @ w /\ u -b-> w).
  Proof.
    intros H; apply ut_beta_inv in H.
    destruct H as [ (f & a & H1 & H2) | [ (f & g & a & H1 & H2 & H3) | [ (f & a & b & H1 & H2 & H3) | (f & g & H1 & H2 & H3) ] ] ]; try discriminate H1;
    apply ut_app_inj in H1; destruct H1; subst h u v.
    left; exists f; auto.
    right; left; exists g; auto.
    right; right; exists b; auto.
  Qed.

  Fact ut_beta_redex_inv n (f : uterm (S n)) (a : uterm n) v :
       \\ f @ a -b-> v ->  v = f[a]
                       \/ (exists g, v = \\ g @ a /\ f -b-> g)
                       \/ (exists b, v = \\ f @ b /\ a -b-> b).
  Proof.
    intros H.
    apply ut_beta_app_inv in H.
    destruct H as [ (g & H & ?) | [ (g & ? & H) | (b & ? & H) ] ]; subst.
    injc H; left; auto.
    apply ut_beta_abs_inv in H.
    destruct H as (f' & ? & H); subst.
    right; left; exists f'; auto.
    right; right; exists b; auto.
  Qed.

  Fact ut_beta_neutral_app_inv n (h u v : uterm n) :
           ut_neutral h  
       ->  h @ u -b-> v 
       -> (exists g, v = g @ u /\ h -b-> g)
       \/ (exists w, v = h @ w /\ u -b-> w).
  Proof.
    intros H1 H2.
    apply ut_beta_app_inv in H2.
    destruct H2 as [ (g & H & ?) | [ (g & ? & H) | (b & ? & H) ] ]; subst.
    destruct H1.
    left; exists g; auto.
    right; exists b; auto.
  Qed.

  Fact ut_beta_rt_abs_inv n f (h : uterm n) : \\ f -b>> h -> exists g, h = \\ g /\ f -b>> g.
  Proof.
    intros H.
    apply clos_refl_trans_map_rev with (f := @ut_abs _) (R := @ut_beta _) in H; auto.
    clear f h H.
    intros ? ? H.
    apply ut_beta_abs_inv in H.
    destruct H as (w & ? & ?); exists w; split; auto; constructor 1; auto.
  Qed.

  Fact ut_beta_var_app_rev_inv n p (v : uterm n) l : 
        £ p *@ l -b-> v -> exists m, v = £ p *@ m /\ Forall2 (inv (@ut_beta_rt _)) m l.
  Proof.
    revert v; induction l as [ | x l IHl ]; simpl; intros v Hv.
    exfalso; revert Hv; apply ut_beta_var_inv.
    apply ut_beta_app_inv in Hv.
    destruct Hv as [ ( g' & ? & ?) | [ (g' & H1 & H2) | (a' & H1 & H2) ] ]; subst.
    destruct l; discriminate H.
    apply IHl in H2.
    destruct H2 as (m & H1 & H2); subst.
    exists (x::m); split; auto; constructor; auto; constructor 2.
    exists (a'::l); split; auto; constructor; auto.
    constructor 1; auto.
    apply Forall2_with_refl, Forall_forall.
    intros; apply ut_beta_rt_refl.
  Qed.

  Fact ut_beta_neutral_app_rev_inv n a (v : uterm n) ll : 
        ut_neutral a
     -> a *@ ll -b-> v -> (exists a', v = a' *@ ll /\ a -b-> a')
                       \/  exists l c c' r, ll = l++c::r /\ v = a *@ (l++c'::r) /\ c -b-> c'.
  Proof.
    revert a v; induction ll as [ | b ll IHll ]; simpl; intros a v H1 H2.
    left; exists v; auto.

    apply ut_beta_app_inv in H2.
    destruct H2 as [ (f & H2 & H3) 
                 | [ (f & H2 & H3)
                   | (f & H2 & H3) ] ]; subst.

    destruct ll; simpl in H2.
    subst a; destruct H1.
    discriminate H2.

    destruct (IHll _ _ H1 H3) as [ (a' & H4 & H5) | (l & c & c' & r & H4 & H5 & H6) ]; subst.
    left; exists a'; auto.
    right; exists (b::l), c, c', r; auto.
    
    right; exists nil, b, f, ll; simpl; auto.
  Qed.

  Fact ut_beta_neutral_app_rec_inv n a b (v : uterm n) ll : 
       (a @ b) @* ll -b-> v -> (exists c,        v = c @* ll 
                                              /\ a @ b  -b-> c)
                            \/  exists l c c' r, ll = l++c::r 
                                              /\ v = (a @ b) @* (l++c'::r) 
                                              /\ c -b-> c'.
  Proof.
    intros H.
    rewrite ut_app_rec_rev_eq in H.
    apply ut_beta_neutral_app_rev_inv in H.
    destruct H as [ (c & ? & H) | (l & c & c' & r & H1 & H2 & H3) ]; subst.
 
    left; exists c; rewrite ut_app_rec_rev_eq; auto.
    right. exists (rev r), c, c', (rev l); repeat split; auto.
    rewrite <- (rev_involutive ll), H1, rev_app_distr; simpl.
    rewrite app_ass; reflexivity.
    rewrite ut_app_rec_rev_eq, rev_app_distr; simpl.
    do 2 rewrite rev_involutive; rewrite app_ass; auto.
    red; auto.
  Qed.

  Fact ut_beta_rt_var_app_rev_inv n p (v : uterm n) l : 
         £ p *@ l -b>> v -> exists m, v = £ p *@ m /\ Forall2 (inv (@ut_beta_rt _)) m l.
  Proof.
    intros H.
    apply clos_refl_trans_map_rev with (f := fun ll => £ p *@ ll) (R := Forall2 (@ut_beta_rt _)) in H; auto.
    destruct H as (m & H1 & H2).
    apply clos_refl_trans_Forall2 in H2.
    exists m; split; auto.
    apply Forall2_sym.
    revert H2; apply Forall2_incl, clos_refl_trans_idem.
    clear v l H.
    intros x z H.
    apply ut_beta_var_app_rev_inv in H.
    destruct H as (m & H1 & H2).
    exists m; split; auto; constructor 1; auto.
    apply Forall2_sym; auto.
  Qed.

  Fact ut_beta_rt_var_app_rec_inv n p (v : uterm n) l : 
         £ p @* l -b>> v -> exists m, v = £ p @* m /\ Forall2 (inv (@ut_beta_rt _)) m l.
  Proof.
    intros H.
    rewrite ut_app_rec_rev_eq in H.
    apply ut_beta_rt_var_app_rev_inv in H.
    destruct H as (m & H1 & H2).
    exists (rev m); split.
    rewrite <- ut_app_rev_rec_eq; auto.
    apply Forall2_rev in H2.
    rewrite rev_involutive in H2.
    auto.
  Qed.

End ut_beta_inv.

Section ut_beta_ren_rel_square.

  Fact ut_beta_ren_rel_comm n m (R : ren_rel n m) : 
       square (@ut_beta _) (ut_ren_rel R) (ut_ren_rel R) (@ut_beta _).
  Proof.
    intros f g u H1.
    revert H1 m R u.
    induction 1 as [ n f a | n f g a H1 IH1 | n a b f H1 IH1 | n f g H1 IH1 ];
      intros m R u H.

    apply ut_ren_rel_app_inv in H.
    destruct H as (h & b & ? & H1 & H2); subst.
    apply ut_ren_rel_abs_inv in H1.
    destruct H1 as (g & ? & H1); subst.
    exists (g[b]); split.
    apply ut_ren_rel_subst with (1 := H1).
    intros p q; pos_inv p; pos_inv q; try (destruct 1; fail).
    intros _; apply H2.
    unfold ren_rel_lift, uts_first; simpl.
    constructor; auto.
    constructor; auto.

    apply ut_ren_rel_app_inv in H.
    destruct H as (h & b & ? & H2 & H3); subst.
    destruct (IH1 _ _ _ H2) as (v & ? & ?).
    exists (v @ b); split.
    constructor; auto.
    constructor 2; auto.

    apply ut_ren_rel_app_inv in H.
    destruct H as (g & c & ? & H2 & H3); subst.
    destruct (IH1 _ _ _ H3) as (v & ? & ?).
    exists (g @ v); split.
    constructor; auto.
    constructor 3; auto.

    apply ut_ren_rel_abs_inv in H.
    destruct H as (h & ? & H); subst.
    destruct (IH1 _ _ _ H) as (v & ? & ?).
    exists (\\ v); split.
    constructor; auto.
    constructor; auto.
  Qed.

End ut_beta_ren_rel_square.

Notation ut_beta_sn := (Acc (inv (@ut_beta _))).

Fact ut_beta_beta_sn n (f g : uterm n) : f -b-> g -> ut_beta_sn f -> ut_beta_sn g.
Proof.
  intros H1 H2; apply (Acc_inv H2 _ H1).
Qed.

Fact ut_beta_beta_rt_sn n (f g : uterm n) : f -b>> g -> ut_beta_sn f -> ut_beta_sn g.
Proof.
  induction 1; auto; apply ut_beta_beta_sn; auto.
Qed.

Corollary ut_beta_ren_sn n m r f : ut_beta_sn f <-> ut_beta_sn (@ut_ren n f m r).
Proof.
  split.

  apply square_Acc with (E := ut_ren_rel (fun p q => p = r q)).
  apply ut_beta_ren_rel_comm.
  apply ut_ren_rel_sym, ut_ren_rel_eq_ren.

  apply square_Acc with (E := ut_ren_rel (fun p q => q = r p)).
  apply ut_beta_ren_rel_comm.
  apply ut_ren_rel_eq_ren.
Qed.

Corollary ut_beta_lift_sn n f : ut_beta_sn f <-> ut_beta_sn (@ut_lift n f).
Proof.
  apply ut_beta_ren_sn.
Qed.

Corollary ut_beta_ren_inv n m f g (r : pos_map n m) : ut_ren f r -b-> g -> exists g', g = ut_ren g' r /\ f -b-> g'.
Proof.
  intros H1.
  generalize (ut_ren_rel_eq_ren f r); intros H2.
  apply ut_ren_rel_sym in H2.
  destruct (ut_beta_ren_rel_comm H1 H2) as (g' & H3 & H4).
  exists g'; split; auto.
  apply ut_ren_rel_fun_eq, ut_ren_rel_sym; auto.
Qed.

Fact ut_beta_rt_ren_inv n m f g (r : pos_map n m) : ut_ren f r -b>> g -> exists g', g = ut_ren g' r /\ f -b>> g'.
Proof.
  gen_eq (ut_ren f r); intros u Hu H; revert H f Hu.
  induction 1 as [ u g H | | u v w H1 IH1 H2 IH2 ]; intros f ?; subst.
  apply ut_beta_ren_inv in H; destruct H as (g' & ? & ?); exists g'; split; auto; constructor 1; auto.
  exists f; split; auto; constructor 2.
  destruct (IH1 _ eq_refl) as (g1 & ? & ?); subst.
  destruct (IH2 _ eq_refl) as (g2 & ? & ?); subst.
  exists g2; split; auto; constructor 3 with g1; auto.
Qed.
  
Corollary ut_beta_lift_inv n (f : uterm n) g : ut_lift f -b-> g -> exists g', g = ut_lift g' /\ f -b-> g'.
Proof.
  apply ut_beta_ren_inv.
Qed.

Corollary ut_beta_rt_lift_inv n (f : uterm n) g : ut_lift f -b>> g -> exists g', g = ut_lift g' /\ f -b>> g'.
Proof.
  apply ut_beta_rt_ren_inv.
Qed.

Local Fact ut_beta_sn_abs_inv n (f : uterm n) : ut_beta_sn f -> forall g, f = \\ g -> ut_beta_sn g.
Proof.
  induction 1 as [ f' Hf IHf ]; intros f ?; subst.
  constructor; intros g Hg.
  apply IHf with (2 := eq_refl).
  constructor; auto.
Qed.

Fact ut_beta_sn_abs n (f : uterm (S n)) : ut_beta_sn f <-> ut_beta_sn (\\ f).
Proof.
  split.

  induction 1 as [ f Hf IH ].
  constructor 1; intros u Hu.
  apply ut_beta_abs_inv in Hu.
  destruct Hu as (g & ? & Hg); subst.
  apply IH, Hg.

  intros H.
  apply ut_beta_sn_abs_inv with (1 := H); reflexivity.
Qed.

Fact ut_beta_sn_abs_rec k n (f : uterm (k+n)) : ut_beta_sn f <-> ut_beta_sn (\\* f).
Proof.
  revert f; induction k as [ | k IHk ]; simpl; intros f.
  tauto.
  rewrite <- IHk.
  apply ut_beta_sn_abs.
Qed.

Fact ut_beta_sn_left_inv n (a b : uterm n) : ut_beta_sn (a @ b) -> ut_beta_sn a.
Proof.
  gen_eq (a @ b).
  intros f E H; revert f H a b E.
  induction 1 as [ f Hf IHf ]; intros a b ?; subst.
  constructor 1; intros a' Ha'.
  apply IHf with (b := b) (2 := eq_refl).
  constructor 2; auto.
Qed.

Fact ut_beta_sn_right_inv n (a b : uterm n) : ut_beta_sn (a @ b) -> ut_beta_sn b.
Proof.
  gen_eq (a @ b).
  intros f E H; revert f H a b E.
  induction 1 as [ f Hf IHf ]; intros a b ?; subst.
  constructor 1; intros b' Hb'.
  apply IHf with (a := a) (2 := eq_refl).
  constructor 3; auto.
Qed.

Fact ut_beta_sn_app_neutral n (f a : uterm n) : 
        (forall f', f -b>> f' -> ut_neutral f') 
      -> ut_beta_sn f 
      -> ut_beta_sn a
      -> ut_beta_sn (f @ a).
Proof.
  intros H1 H2 H3.
  revert H2 a H3 H1.
  induction 1 as [ f Hf IHf ].
  induction 1 as [ a Ha IHa ].
  intros H.
  constructor 1; intros v Hv.
  apply ut_beta_neutral_app_inv in Hv.
  2: apply H; constructor 2.
  destruct Hv as [ (g & ? & Hg) | (b & ? & Hb) ]; subst.
  apply (IHf _ Hg); auto.
  constructor; apply Ha.
  intros h Hh; apply H.
  constructor 3 with g; auto; constructor 1; auto.
  apply (IHa _ Hb); auto.
Qed.

Section uterm_beta_inv_ren.

  Variables (n m : nat) (r : pos_map n m) (Hr : forall p q, r p = r q -> p = q).
  
  Let R x y := y = r x.
  
  Let HR x u v : R u x -> R v x -> u = v.
  Proof.
    unfold R; intros; subst; apply Hr; auto.
  Qed.
  
  Let HR0 f : forall p, 0 < ut_occ f p -> (R o inv R) p p.
  Proof. intros p; exists (r p); split; reflexivity. Qed.
  
  Let HR1 f : ut_ren_rel (inv R) (ut_ren f r) f.
  Proof.
    generalize (@ut_ren_rel_id _ f _ (HR0 f)).
    intros H.
    apply ut_ren_rel_comp in H.
    destruct H as (g & H1 & H2).
    apply ut_ren_rel_fun_eq in H1; subst; auto.
  Qed.

  Fact ut_beta_inv_ren f g : ut_ren f r -b-> ut_ren g r -> f -b-> g.
  Proof.
    generalize (HR1 f) (HR1 g).
    apply ut_beta_ren_rel_fun; auto.
  Qed.
  
End uterm_beta_inv_ren.
 
Fact ut_beta_inv_lift n (f g : uterm n) : ut_lift f -b-> ut_lift g -> f -b-> g.
Proof.
  apply ut_beta_inv_ren, pos_nxt_inj.
Qed.

(* beta reduction has a finite direct image *)

Fact ut_beta_finite n (u : uterm n) : { l | forall v, u -b-> v <-> In v l }. 
Proof.
  induction u as [ n f (l & Hl) | n a (la & Ha) b (lb & Hb) | ].

  (* when u is a lambda \\ f, reduction must occur below the lambda in f *)

  exists (map (@ut_abs _) l); intros v; split; intros H.
  apply ut_beta_abs_inv in H; destruct H as (w & ? & H); subst.
  apply in_map_iff; exists w; split; auto.
  apply Hl; auto.

  apply in_map_iff in H.
  destruct H as (g & ? & H); subst.
  constructor; apply Hl; auto.

  (* when u is an app a @ b, either one of the subterm (a or b) reduces 
     or we have a redex, ie a is \\ f for some f *)
  
  exists (map (fun x => x @ b) la ++ map (fun x => a @ x) lb
          ++ match a with \\ f => f [b]::nil | _ => nil end).
  intros v; split; intros H.

  apply ut_beta_app_inv in H.
  destruct H as [ (f & ? & H) | [ (c & ? & H) | (c & ? & H) ] ]; subst.
  do 2 (apply in_or_app; right); left; auto.
  apply in_or_app; left.
  apply in_map_iff; exists c; split; auto.
  apply Ha; auto.
  apply in_or_app; right; apply in_or_app; left.
  apply in_map_iff; exists c; split; auto.
  apply Hb; auto.

  apply in_app_or in H; destruct H as [ H | H ].
  apply in_map_iff in H.
  destruct H as (c & ? & H); subst.
  constructor 2; apply Ha; auto.
  apply in_app_or in H; destruct H as [ H | H ].
  apply in_map_iff in H.
  destruct H as (c & ? & H); subst.
  constructor 3; apply Hb; auto.
  destruct a as [ ? f | | ]; try (destruct H; fail).
  destruct H as [ [] | [] ]; constructor 1.

  (* a variable does not reduce to anything *)

  exists nil; split.
  apply ut_beta_var_inv.
  intros [].
Qed.

Section implementation_of_SN.

  Variable (n : nat).
  
  Implicit Type (u : uterm n).

  (* Hence the direct image of beta reduction is computed into a list *)

  Definition ut_beta_list u := proj1_sig (@ut_beta_finite n u).

  Definition ut_beta_list_prop u : forall v, u -b-> v <-> In v (ut_beta_list u).
  Proof. apply (proj2_sig (@ut_beta_finite n u)). Qed.

  (* Strongly normalising terms are those which are accessible under the
     inverse of beta reduction *) 

  Notation SN := ut_beta_sn.

  (* Property 1 is for free *)

  Fact ut_beta_sn_intro u : (forall v, u -b-> v -> SN v) -> SN u.
  Proof. apply Acc_intro with (R := inv (@ut_beta _)). Qed.

  (* there is a path u -b-> ... -b-> v of length p *)

  Notation "u '-[' p ']->' v" := (rel_iter (fun x y => x -b-> y) p u v) (at level 70).

  (* The height (maximum length of a reduction) can be computed from the
     SN/accessibility property, see finitary_rel.v for the generic
     implementation *)

  Definition ut_beta_sn_ht u : SN u -> nat.
  Proof. 
    apply fin_rel_ht with (f := ut_beta_list).
    intros; apply ut_beta_list_prop; tauto.
  Defined.

  (* the height needs a proof of SN/accessibility but the result does not depend on it *)

  Fact ut_beta_sn_ht_eq u (H1 H2 : SN u) : ut_beta_sn_ht H1 = ut_beta_sn_ht H2.
  Proof. apply fin_rel_ht_eq. Qed.

  (* the height is an upper bound *)

  Fact ut_beta_sn_ht_bound u (Hu : SN u) p v : u -[p]-> v -> p <= ut_beta_sn_ht Hu.
  Proof. apply fin_rel_ht_bound. Qed.

  (* and this upper bound is reached for some computable term *)

  Fact ut_beta_sn_ht_lub u (Hu : SN u) : { v | u -[ut_beta_sn_ht Hu]-> v }.
  Proof. apply fin_rel_ht_lub. Qed.

  (* We deduce Property 2 *)

  Fact ut_beta_sn_dec u (Hu : SN u) v (Hv : SN v) : 
      u -b-> v -> ut_beta_sn_ht Hv < ut_beta_sn_ht Hu.
  Proof.
    intros H.
    destruct (ut_beta_sn_ht_lub Hv) as (w & Hw).
    apply ut_beta_sn_ht_bound with w.
    exists v; auto.
  Qed.

  Fact ut_beta_rt_sn_dec u (Hu : SN u) v (Hv : SN v) : 
      u -b>> v -> ut_beta_sn_ht Hv <= ut_beta_sn_ht Hu.
  Proof.
    intros H; revert H Hu Hv.
    induction 1 as [ | | u v w H1 IH1 H2 IH2 ]; intros Hu Hv.
    apply lt_le_weak, ut_beta_sn_dec; auto.
    rewrite (ut_beta_sn_ht_eq Hu Hv); auto.
    apply le_trans with (2 := IH1 Hu (ut_beta_beta_rt_sn H1 Hu)); auto.
  Qed.

End implementation_of_SN.

Implicit Arguments ut_beta_sn_ht [ n ].

Definition ut_beta_normal_form n f : Acc (inv (@ut_beta n)) f -> { g | f -b>> g /\ forall h, ~ g -b-> h }.
Proof.
  apply finitary_normal_form, ut_beta_finite.
Qed.

Definition ut_sn_ht_rel n m f g :=
  exists af ag, @ut_beta_sn_ht n f af < @ut_beta_sn_ht m g ag
             \/ ut_beta_sn_ht f af = ut_beta_sn_ht g ag /\ ut_height f < ut_height g.

Fact ut_sn_ht_rel_sn n m f g : @ut_sn_ht_rel n m f g -> ut_beta_sn f.
Proof.
  intros (af & ?); auto.
Qed.

Section ut_sn_lexico.

  Let R (c1 c2 : nat * nat) := 
    match c1, c2 with
      | (a1,b1), (a2,b2) => a1 < a2 \/ a1 = a2 /\ b1 < b2
    end.

  Let Rwf : well_founded R.
  Proof.
    intros (a2,b2).
    revert b2; induction a2 as [ a2 IHa ] using (well_founded_induction lt_wf).
    induction b2 as [ b2 IHb ] using (well_founded_induction lt_wf).
    constructor 1.
    intros (a1,b1) [ ? | (H1 & H2) ].
    apply IHa; auto.
    subst a1; apply IHb; auto.
  Qed.

  Section ut_sn_ht_rect.

    Variable P : forall n (f : uterm n), Type.

    Hypothesis HP : forall m g, (forall n f, @ut_sn_ht_rel n m f g -> P f) -> P g.

    Let lemma c : forall n f af, c = (@ut_beta_sn_ht n f af, ut_height f) -> P f. 
    Proof.
      induction c as [ (a,b) IH ] using (well_founded_induction_type Rwf).
      intros n f af H.
      injc H.
      apply HP.
      intros m g Hg.
      apply IH with (y := (ut_beta_sn_ht g (ut_sn_ht_rel_sn Hg), ut_height g)) (2 := eq_refl).
      red.
      destruct Hg as (ag1 & af1 & H ).
      rewrite ut_beta_sn_ht_eq with (H2 := ag1).
      rewrite ut_beta_sn_ht_eq with (H2 := af1); auto.
    Qed.

    Theorem ut_sn_ht_rect n (f : uterm n) : ut_beta_sn f -> P f.
    Proof.
      intros Hf.
      apply lemma with (af := Hf) (1 := eq_refl).
    Qed.

  End ut_sn_ht_rect.

  Section ut_sn_lexico.

    Variable P : nat -> nat -> Type.

    Hypothesis HP : forall c d, (forall a b, R (a,b) (c,d) -> P a b) -> P c d.

    Theorem ut_sn_lexico a b : P a b.
    Proof.
      set (Q c := P (fst c) (snd c)).
      change (Q (a,b)).
      generalize (a,b); clear a b.
      induction p as [ (a,b) IH ] using (well_founded_induction_type Rwf). 
      apply HP. 
      intros c d; simpl. 
      apply (IH (c,d)).
    Qed.
 
  End ut_sn_lexico.

End ut_sn_lexico.

Fact ut_st_beta_sn n m f g : @ut_subterm n m f g -> ut_beta_sn g -> ut_beta_sn f.
Proof.
  induction 1; auto.
  apply ut_beta_sn_abs.
  apply ut_beta_sn_left_inv.
  apply ut_beta_sn_right_inv.
Qed.

Section ut_beta_sn_ht_rel.

  Variable (n m : nat) (R : uterm n -> uterm m -> Prop).

  Hypothesis HR : forall f f' g, R f g -> f -b-> f' -> exists g', R f' g' /\ g -b-> g'.

  Notation "u '-[' p ']->' v" := (rel_iter (fun x y => x -b-> y) p u v) (at level 70).

  Let HR' : forall k f f' g, R f g -> f -[k]-> f' -> exists g', R f' g' /\ g -[k]-> g'.
  Proof.
    induction k as [ | k IHk ]; simpl; intros f f' g H1 H2.
    subst; exists g; auto.
    destruct H2 as (f1 & H2 & H3).
    destruct (HR H1 H2) as (g1 & H4 & H5).
    destruct (IHk _ _ _ H4 H3) as (g' & H6 & H7).
    exists g'; split; auto.
    exists g1; auto.
  Qed.

  Theorem ut_beta_sn_ht_rel f g : R f g -> forall af ag, ut_beta_sn_ht f af <= ut_beta_sn_ht g ag.
  Proof.
    intros H1 af ag.
    destruct (ut_beta_sn_ht_lub af) as (f' & Hf').
    destruct (HR' _ _ H1 Hf') as (g' & H2 & H3).
    apply (ut_beta_sn_ht_bound ag _ _ H3).
  Qed.

End ut_beta_sn_ht_rel.

Fact ut_beta_sn_ht_abs n f H1 H2 : @ut_beta_sn_ht (S n) f H1 = ut_beta_sn_ht (\\ f) H2. 
Proof.
  apply le_antisym.

  revert H1 H2; apply ut_beta_sn_ht_rel with (R := fun u v => v = \\ u); auto.
  clear f; intros f f' g ? H; subst; exists (\\ f'); split; auto.
  constructor; auto.

  revert H2 H1; apply ut_beta_sn_ht_rel with (R := fun u v => u = \\ v); auto.
  clear f; intros f f' g ?; subst.
  apply ut_beta_abs_inv.
Qed.

Fact ut_beta_sn_ht_abs_rec k n (f : uterm (k+n)) H1 H2 : 
   ut_beta_sn_ht f H1 = ut_beta_sn_ht (\\* f) H2.
Proof.
  revert f H1 H2; induction k as [ | k IHk ]; simpl; intros f H1 H2.
  apply ut_beta_sn_ht_eq.
  rewrite <- IHk with (H1 := proj1 (ut_beta_sn_abs _) H1).
  apply ut_beta_sn_ht_abs.
Qed.

Fact ut_beta_sn_ht_left n a b H1 H2 : @ut_beta_sn_ht n a H1 <= ut_beta_sn_ht (a @ b) H2.
Proof.
  revert H1 H2;
  apply ut_beta_sn_ht_rel with (R := fun u v => v = u @ b); auto.
  clear a; intros f f' g ? H; subst.
  exists (f' @ b); split; auto.
  constructor 2; auto.
Qed.

Fact ut_beta_sn_ht_right n a b H1 H2 : @ut_beta_sn_ht n b H1 <= ut_beta_sn_ht (a @ b) H2.
Proof.
  revert H1 H2;
  apply ut_beta_sn_ht_rel with (R := fun u v => v = a @ u); auto.
  clear b; intros f f' g ? H; subst.
  exists (a @ f'); split; auto.
  constructor 3; auto.
Qed.

Fact ut_st_sn_ht n m f g : @ut_subterm n m f g -> forall af ag, ut_beta_sn_ht f af <= ut_beta_sn_ht g ag.
Proof.
  induction 1 as [ | | | n m k a b c H1 IH1 H2 IH2 ]; intros af ag; auto.
  rewrite ut_beta_sn_ht_abs with (H2 := ag); apply le_refl.
  apply ut_beta_sn_ht_left.
  apply ut_beta_sn_ht_right. 
  apply le_trans with (2 := IH2 (ut_st_beta_sn H2 ag) ag); auto.
Qed.

(* This is the induction principle used to show that SN term are typable in system D *)

Section ut_sn_st_rect.
   
  Variable P : forall n (f : uterm n), Type.

  Hypothesis HP : forall m g ag, (forall n f af, 
                    (  @ut_beta_sn_ht n f af < @ut_beta_sn_ht m g ag 
                    \/ @ut_subterm n m f g ) -> P f) -> P g.

  Theorem ut_sn_st_rect n (f : uterm n) : ut_beta_sn f -> P f.
  Proof.
    intros Hf.
    generalize Hf.
    set (Q n (f : uterm n) := ut_beta_sn f -> P f).
    change (Q n f).
    revert n f Hf. 
    apply ut_sn_ht_rect; unfold Q; clear Q.
    intros m g Hg ag.
    apply HP with (ag := ag).
    intros n f af Hf.
    apply Hg; auto.
    exists af, ag.
    destruct Hf as [ Hf | Hf ].
    left; auto.
    generalize (ut_st_sn_ht Hf af ag); intros H.
    destruct (le_lt_dec (ut_beta_sn_ht _ ag) (ut_beta_sn_ht _ af)).
    right; split.
    omega.
    revert Hf.
    apply ut_st_ht.
    left; auto.
  Qed.

End ut_sn_st_rect.

Ltac ut_beta_auto :=
  repeat
  match goal with 
    | |- _ @ ?x     -b-> _ @ ?x  => apply in_ut_beta_lft
    | |- ?f @ _     -b-> ?f @ _  => apply in_ut_beta_rt
    | |- \\ _       -b-> \\ _    => apply in_ut_beta_abs
    | |- \\ ?f @ ?a -b-> ?f[?a]  => apply in_ut_beta_redex
  end.

Ltac ut_beta_rt_step := 
  match goal with
    |- ?x -b>> _ => 
    match x with
      context f [ \\ ?h @ ?g ] => let y := context f [ut_reduce h g] in apply ut_beta_rt_trans with y
    end
  end; [ constructor 1; ut_beta_auto | idtac ].

Ltac ut_simpl :=
  repeat (
  match goal with
    | |- context f [ ut_reduce ]  => unfold ut_reduce
    | |- context f [ uts_lift  ]  => unfold uts_lift
    | |- context f [ ut_lift  ]   => unfold ut_lift
    | |- context f [ uts_first  ] => unfold uts_first
    | |- context f [ ut_ren ]     => rewrite ut_ren_subst_eq
    | |- context f [ pm_lift ]    => unfold pm_lift
  end; simpl);
  repeat rewrite ut_subst_comp; simpl;
  repeat rewrite ut_subst_id.


  
