(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import utils pos.
Require Import uterm uterm_beta uterm_head.

Set Implicit Arguments.

Section ut_specialize.

  Variables (n k : nat) (f : forall i, uterm i -> uterm i) (phi : uterm n).

  Hypothesis Hf_subst : forall i j t s, ut_subst (@f i t) s = @f j (ut_subst t s).
  Hypothesis Hf_whead : forall i u v, u -w>> v -> @f i u -w>> f v.
  Hypothesis Hf_beta : forall i u v, u -b>> v -> @f i u -b>> f v.

  Let ll i := map (fun p => £ (pos_lft n (pos_nxt _ p))) (pos_list i).

  Definition ut_specialize := \\* f (ut_lift_rec (S k) phi @ £ (pos_fst _) *@ ll k).

  Let uts_eq v t : ut_specialize *@ map (pf_cons t (fun p => v(pos_rev p))) (pos_list (S k))
                 = ut_specialize *@ map (fun p : pos k => v (pos_rev p)) (pos_list k) @ t.
  Proof.
    simpl; do 2 f_equal.
    rewrite map_map.
    apply map_ext.
    intros; unfold pf_cons; simpl; auto.
  Qed.

  Fact ut_specialize_wreduce v t : ut_specialize @* map v (pos_list k) @ t 
                            -w>> f (phi @* t::map v (pos_list k)).
  Proof.
    rewrite (ut_app_rec_rev_eq ut_specialize), <- map_rev, pos_list_map_rev.
    rewrite <- uts_eq.
    unfold ut_specialize.
    match goal with 
      | |- \\* ?f *@ map ?s _  -w>> _ => generalize (ut_abs_rec_wreduce f s)
    end; intros H1; apply ut_whead_rt_trans with (1 := H1); clear H1.
    rewrite Hf_subst.
    rewrite ut_subst_app_rec_rev.
    unfold ut_subst; fold ut_subst.
    rewrite ut_subst_lift_rec, ut_subst_id.
    rewrite pf_merge_pos_fst, pf_cons_fst_eq.
    apply Hf_whead.
    simpl.
    rewrite ut_app_rev_rec_eq.
    apply ut_eq_whead_rt.
    f_equal.
    unfold ll.
    do 2 rewrite <- map_rev.
    rewrite map_map, pos_list_map_rev.
    apply map_ext.
    intro; simpl.
    rewrite pf_merge_pos_nxt, pf_merge_lft_eq; simpl.
    rewrite pf_cons_nxt_eq, pos_rev_rev.
    auto.
  Qed.

  Fact ut_specialize_breduce v t : ut_specialize @* map v (pos_list k) @ t 
                            -b>> f (phi @* t::map v (pos_list k)).
  Proof.
    rewrite (ut_app_rec_rev_eq ut_specialize), <- map_rev, pos_list_map_rev.
    rewrite <- uts_eq.
    unfold ut_specialize.
    match goal with 
      | |- \\* ?f *@ map ?s _  -b>> _ => generalize (ut_abs_rec_reduce f s)
    end; intros H1; apply ut_beta_rt_trans with (1 := H1); clear H1.
    rewrite Hf_subst.
    rewrite ut_subst_app_rec_rev.
    unfold ut_subst; fold ut_subst.
    rewrite ut_subst_lift_rec, ut_subst_id.
    rewrite pf_merge_pos_fst, pf_cons_fst_eq.
    apply Hf_beta.
    simpl.
    rewrite ut_app_rev_rec_eq.
    apply ut_eq_beta_rt.
    f_equal.
    unfold ll.
    do 2 rewrite <- map_rev.
    rewrite map_map, pos_list_map_rev.
    apply map_ext.
    intro; simpl.
    rewrite pf_merge_pos_nxt, pf_merge_lft_eq; simpl.
    rewrite pf_cons_nxt_eq, pos_rev_rev.
    auto.
  Qed.

End ut_specialize.

